//
// GameX - Miscellaneous Interface-related WindowsDX Class Code
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source License. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 
// this is a version of GameX that has been modified (by Justin Pease)


#include "gamex-win-dx.hpp"




void WindowsDX::InfoDialog(char * message)
{
	SpecialPauseFade(0.125f,0.25f,0,0,32,64,64,128,160);
	MessageBox (win_hwnd, message, "Notice", MB_OK|MB_DEFBUTTON1|MB_ICONINFORMATION);
}

void WindowsDX::WarningDialog(char * message)
{
	SpecialPauseFade(0.125f,0.25f,64,128,64,128,0,0,160);
	MessageBox (win_hwnd, message, "Warning", MB_OK|MB_DEFBUTTON1|MB_ICONWARNING);
}

int WindowsDX::ChoiceDialog(char * message, bool three_choices, int default_choice)
{
	SpecialPauseFade(0.125f,0.25f,0,0,64,128,32,64,160);
	int option = three_choices ? MB_YESNOCANCEL : MB_YESNO;
	switch(default_choice) {
		case 1:  option |= MB_DEFBUTTON1; break;
		case 0:  option |= MB_DEFBUTTON2; break;
		case -1: if(three_choices) option |= MB_DEFBUTTON3; break;
	}
	int choice = MessageBox (win_hwnd, message, "Query", option|MB_ICONQUESTION);
	switch(choice) {
	case IDCANCEL:
		return -1;
	case IDNO:
		return 0;
	default:
	case IDYES:
		return 1;
	}
}

void WindowsDX::ErrorDialog(char * message)
{
	SpecialPauseFade(0.125f,0.25f,64,128,0,0,32,64,160);
	MessageBox (win_hwnd, message, "Error", MB_OK|MB_DEFBUTTON1|MB_ICONERROR);
}

int WindowsDX::ErrorChoiceDialog(char * message, int default_choice)
{
	SpecialPauseFade(0.125f,0.25f,64,128,0,0,32,64,160);
	int option = MB_ABORTRETRYIGNORE;
	switch(default_choice) {
		case -1:  option |= MB_DEFBUTTON1; break;
		case 1:  option |= MB_DEFBUTTON2; break;
		case 0: option |= MB_DEFBUTTON3; break;
	}
	int choice = MessageBox (win_hwnd, message, "Error", option|MB_ICONERROR);
	switch(choice) {
	case IDABORT:
		return -1;
	case IDIGNORE:
		return 0;
	default:
	case IDRETRY:
		return 1;
	}
}

bool WindowsDX::CheckQuit(void)
{
	if(win_request_no_confirm_quit) { // if quit confirmation is off
		quit_dialog = false;
		return true; // signal to exit immediately
	}
	else {

		quit_dialog = true;
		int choice = 0;

		bool hidden_cursor = win_cursor_hidden;
		if(hidden_cursor) {
			UnhideCursor();
			win_request_no_hide_cursor = true; // avoids re-hiding
		}

		choice = MessageBox (win_hwnd, "Are you sure you want to exit?", GameX.GetName(), MB_YESNO|MB_DEFBUTTON1|MB_ICONQUESTION);

		if(hidden_cursor) {
			win_request_no_hide_cursor = false; // reset to hide
			HideCursor();
		}

		quit_dialog = false;

		return (choice == IDYES); // if true, the game quits, if false, the game keeps going
	}
}

// The game can call this to exit
void WindowsDX::Quit (bool unrecoverable)
{
	if(unrecoverable) { // immediately quit if it's a fatal-error-type quit
		if(win_log_debug)
			debug.Output("WinDX: Exiting immediately...\n");
		GameX.CleanUp() ;
		exit(-1);
	}

	if(!(win_request_no_confirm_quit)) {
		quit_dialog = true;
#ifndef GAMEX_RUN_UNMANAGED
		SetWarningOnDraw(false);
		GameDraw();
#endif

		SpecialPauseFade(0.125f,0.25f,64,128,32,64,0,0,160);
		UpdateView();
	}

	SendMessage(win_hwnd,WM_CLOSE,0,0); // SendMessage instead of PostMessage, for immediate action
}


void WindowsDX::SetApplicationIcon(int idi_icon_id)
{
	win_appIconIDStr = MAKEINTRESOURCE(idi_icon_id);
}


void WindowsDX::SetMouseMove (int x, int y)
{
	mouse_x = x;
	mouse_y = y;

	if(!win_fullscreen && win_active && CanHideCursor()) {
		int dx = 2*GetMouseDX();
		int dy = 2*GetMouseDY();
		if(x < 0 || y < 0 || x >= win_realx || y >= win_realy) {
			UnhideCursor();
		} else {
			HideCursor();
		}
	}
}

void WindowsDX::HideCursor(void)
{
	if(!win_cursor_hidden) {
		ShowCursor(false);
		win_cursor_hidden = true;
	}
}

void WindowsDX::UnhideCursor(void)
{
	if(win_cursor_hidden) {
		ShowCursor(true);
		win_cursor_hidden = false;
	}
}


void WindowsDX::SetMousePosition (int x, int y)
{
	if(!di_mousestop) {
		mouse_x = x*win_realx/win_width;
		mouse_y = y*win_realy/win_height;
		SetCursorPos(win_xpos+win_xoffset+mouse_x, win_ypos+win_yoffset+mouse_y);
	}
}




void WindowsDX::SetPosition (int x, int y)
{
	if(x > win_screenx) x -= 65535; // reverse unsigning to avoid totally wrong window position
	if(y > win_screeny) y -= 65535;

	win_xpos = x;
	win_ypos = y;
}

void WindowsDX::SetProportional (void)
{
	int screenx = win_realx;
	int screeny = win_realy;

	if(win_width * screeny > win_height * screenx) {
		win_realx = win_width*screenx/win_width;
		win_realy = win_height*screenx/win_width;
	} else {
		win_realx = win_width*screeny/win_height;
		win_realy = win_height*screeny/win_height;
	}

	win_xoffset = (screenx - win_realx)/2; // temporarily acts as initial offset, which doesn't change
	win_yoffset = (screeny - win_realy)/2; // for fullscreen mode, allowing the viewport to be centered
}

void WindowsDX::SetView (int left, int top, int right, int bottom)
{
	win_viewx1 = left;
	win_viewy1 = top;
	win_viewx2 = right;
	win_viewy2 = bottom;
}

void WindowsDX::ResetView (void)
{
	SetView (0, 0, win_width, win_height);
}

void WindowsDX::GetView (int& xLeft, int& yTop, int& xRight, int& yBottom)
{
	 xLeft = win_viewx1;
	 yTop = win_viewy1;
	 xRight = win_viewx2;
	 yBottom = win_viewy2;
}

void WindowsDX::DirectDrawError (HRESULT status)
{
	if(SUCCEEDED(status) || status == dd_last_error) return;
	dd_last_error = status;
	switch (status) {
	case DDERR_ALREADYINITIALIZED: debug.Output ("DirectDraw: This object is already initialized.\n") ; break;
	case DDERR_BLTFASTCANTCLIP: debug.Output ("DirectDraw: Return if a clipper object is attached to the source surface passed into a BltFast call.\n") ; break;
	case DDERR_CANNOTATTACHSURFACE: debug.Output ("DirectDraw: This surface can not be attached to the requested surface.\n") ; break; 
	case DDERR_CANNOTDETACHSURFACE: debug.Output ("DirectDraw: This surface can not be detached from the requested surface.\n") ; break;
	case DDERR_CANTCREATEDC: debug.Output ("DirectDraw: Windows can not create any more DCs\n") ; break;
	case DDERR_CANTDUPLICATE: debug.Output ("DirectDraw: Can't duplicate primary & 3D surfaces, or surfaces that are implicitly created.\n") ; break;
	case DDERR_CLIPPERISUSINGHWND: debug.Output ("DirectDraw: An attempt was made to set a cliplist for a clipper object that is already monitoring an hwnd.\n") ; break;
	case DDERR_COLORKEYNOTSET: debug.Output ("DirectDraw: No src color key specified for this operation.\n") ; break;
	case DDERR_CURRENTLYNOTAVAIL: debug.Output ("DirectDraw: Support is currently not available.\n") ; break;
	case DDERR_DIRECTDRAWALREADYCREATED: debug.Output ("DirectDraw: A DirectDraw object representing this driver has already been created for this process.\n") ; break;
	case DDERR_EXCEPTION: debug.Output ("DirectDraw: An exception was encountered while performing the requested operation.\n") ; break;
	case DDERR_EXCLUSIVEMODEALREADYSET: debug.Output ("DirectDraw: An attempt was made to set the cooperative level when it was already set to exclusive.\n") ; break;
	case DDERR_GENERIC: debug.Output ("DirectDraw: Generic failure. Unfortunately, this means that DirectX has no idea what went wrong.\n") ; break;
	case DDERR_HEIGHTALIGN: debug.Output ("DirectDraw: Height of rectangle provided is not a multiple of reqd alignment.\n") ; break;
	case DDERR_HWNDALREADYSET: debug.Output ("DirectDraw: The CooperativeLevel HWND has already been set. It can not be reset while the process has surfaces or palettes created.\n") ; break;
	case DDERR_HWNDSUBCLASSED: debug.Output ("DirectDraw: HWND used by DirectDraw CooperativeLevel has been subclassed, this prevents DirectDraw from restoring state.\n") ; break;
	case DDERR_IMPLICITLYCREATED: debug.Output ("DirectDraw: This surface can not be restored because it is an implicitly created surface.\n") ; break;
	case DDERR_INCOMPATIBLEPRIMARY: debug.Output ("DirectDraw: Unable to match primary surface creation request with existing primary surface.\n") ; break;
	case DDERR_INVALIDCAPS: debug.Output ("DirectDraw: One or more of the caps bits passed to the callback are incorrect.\n") ; break;
	case DDERR_INVALIDCLIPLIST: debug.Output ("DirectDraw: DirectDraw does not support the provided cliplist.\n") ; break;
	case DDERR_INVALIDDIRECTDRAWGUID: debug.Output ("DirectDraw: The GUID passed to DirectDrawCreate is not a valid DirectDraw driver identifier.\n") ; break;
	case DDERR_INVALIDMODE: debug.Output ("DirectDraw: DirectDraw does not support the requested mode.\n") ; break;
	case DDERR_INVALIDOBJECT: debug.Output ("DirectDraw: DirectDraw received a pointer that was an invalid DIRECTDRAW object.\n") ; break;
	case DDERR_INVALIDPARAMS: debug.Output ("DirectDraw: One or more of the parameters passed to the function are incorrect.\n") ; break;
	case DDERR_INVALIDPIXELFORMAT: debug.Output ("DirectDraw: The pixel format was invalid as specified.\n") ; break;
	case DDERR_INVALIDPOSITION: debug.Output ("DirectDraw: Returned when the position of the overlay on the destination is no longer legal for that destination->\n") ; break;
	case DDERR_INVALIDRECT: debug.Output ("DirectDraw: Rectangle provided was invalid.\n") ; break;
	case DDERR_LOCKEDSURFACES: debug.Output ("DirectDraw: Operation could not be carried out because one or more surfaces are locked.\n") ; break;
	case DDERR_NO3D: debug.Output ("DirectDraw: There is no 3D present.\n") ; break;
	case DDERR_NOALPHAHW: debug.Output ("DirectDraw: Operation could not be carried out because there is no alpha accleration hardware present or available.\n") ; break;
	//case DDERR_NOANTITEARHW: debug.Output ("DirectDraw: Operation could not be carried out because there is no hardware support for synchronizing blits to avoid tearing.	\n") ; break;
	case DDERR_NOBLTHW: debug.Output ("DirectDraw: No blitter hardware present.\n") ; break;
	//case DDERR_NOBLTQUEUEHW: debug.Output ("DirectDraw: Operation could not be carried out because there is no hardware support for asynchronous blitting.\n") ; break;
	case DDERR_NOCLIPLIST: debug.Output ("DirectDraw: No cliplist available.\n") ; break;
	case DDERR_NOCLIPPERATTACHED: debug.Output ("DirectDraw: No clipper object attached to surface object.\n") ; break;
	case DDERR_NOCOLORCONVHW: debug.Output ("DirectDraw: Operation could not be carried out because there is no color conversion hardware present or available.\n") ; break;
	case DDERR_NOCOLORKEY: debug.Output ("DirectDraw: Surface doesn't currently have a color key\n") ; break;
	case DDERR_NOCOLORKEYHW: debug.Output ("DirectDraw: Operation could not be carried out because there is no hardware support of the destination color key.\n") ; break;
	case DDERR_NOCOOPERATIVELEVELSET: debug.Output ("DirectDraw: Create function called without DirectDraw object method SetCooperativeLevel being called.\n") ; break;
	case DDERR_NODC: debug.Output ("DirectDraw: No DC was ever created for this surface.\n") ; break;
	case DDERR_NODDROPSHW: debug.Output ("DirectDraw: No DirectDraw ROP hardware.\n") ; break;
	case DDERR_NODIRECTDRAWHW: debug.Output ("DirectDraw: A hardware-only DirectDraw object creation was attempted but the driver did not support any hardware.\n") ; break;
	case DDERR_NOEMULATION: debug.Output ("DirectDraw: Software emulation not available.\n") ; break;
	case DDERR_NOEXCLUSIVEMODE: debug.Output ("DirectDraw: Operation requires the application to have exclusive mode but the application does not have exclusive mode.\n") ; break;
	case DDERR_NOFLIPHW: debug.Output ("DirectDraw: Flipping visible surfaces is not supported.\n") ; break;
	case DDERR_NOGDI: debug.Output ("DirectDraw: There is no GDI present.\n") ; break;
	case DDERR_NOHWND: debug.Output ("DirectDraw: Clipper notification requires an HWND or no HWND has previously been set as the CooperativeLevel HWND.\n") ; break;
	case DDERR_NOMIRRORHW: debug.Output ("DirectDraw: Operation could not be carried out because there is no hardware present or available.\n") ; break;
	case DDERR_NOOVERLAYDEST: debug.Output ("DirectDraw: Returned when GetOverlayPosition is called on an overlay that UpdateOverlay has never been called on to establish a destination->\n") ; break;
	case DDERR_NOOVERLAYHW: debug.Output ("DirectDraw: Operation could not be carried out because there is no overlay hardware present or available.\n") ; break;
	case DDERR_NOPALETTEATTACHED: debug.Output ("DirectDraw: No palette object attached to this surface.\n") ; break;
	case DDERR_NOPALETTEHW: debug.Output ("DirectDraw: No hardware support for 16 or 256 color palettes.\n") ; break;
	case DDERR_NORASTEROPHW: debug.Output ("DirectDraw: Operation could not be carried out because there is no appropriate raster op hardware present or available.\n") ; break;
	case DDERR_NOROTATIONHW: debug.Output ("DirectDraw: Operation could not be carried out because there is no rotation hardware present or available.\n") ; break;
	case DDERR_NOSTRETCHHW: debug.Output ("DirectDraw: Operation could not be carried out because there is no hardware support for stretching.\n") ; break;
	case DDERR_NOT4BITCOLOR: debug.Output ("DirectDraw: DirectDrawSurface is not in 4 bit color palette and the requested operation requires 4 bit color palette.\n") ; break;
	case DDERR_NOT4BITCOLORINDEX: debug.Output ("DirectDraw: DirectDrawSurface is not in 4 bit color index palette and the requested operation requires 4 bit color index palette.\n") ; break;
	case DDERR_NOT8BITCOLOR: debug.Output ("DirectDraw: DirectDrawSurface is not in 8 bit color mode and the requested operation requires 8 bit color.\n") ; break;
	case DDERR_NOTAOVERLAYSURFACE: debug.Output ("DirctX: Returned when an overlay member is called for a non-overlay surface.\n") ; break;
	case DDERR_NOTEXTUREHW: debug.Output ("DirectDraw: Operation could not be carried out because there is no texture mapping hardware present or available.\n") ; break;
	case DDERR_NOTFLIPPABLE: debug.Output ("DirectDraw: An attempt has been made to flip a surface that is not flippable.\n") ; break;
	case DDERR_NOTFOUND: debug.Output ("DirectDraw: Requested item was not found.\n") ; break;
	case DDERR_NOTLOCKED: debug.Output ("DirectDraw: Surface was not locked.  An attempt to unlock a surface that was not locked at all, or by this process, has been attempted.\n") ; break;
	case DDERR_NOTPALETTIZED: debug.Output ("DirectDraw: The surface being used is not a palette-based surface.\n") ; break;
	case DDERR_NOVSYNCHW: debug.Output ("DirectDraw: Operation could not be carried out because there is no hardware support for vertical blank synchronized operations.\n") ; break;
	case DDERR_NOZBUFFERHW: debug.Output ("DirectDraw: Operation could not be carried out because there is no hardware support for zbuffer blitting.\n") ; break;
	case DDERR_NOZOVERLAYHW: debug.Output ("DirectDraw: Overlay surfaces could not be z layered based on their BltOrder because the hardware does not support z layering of overlays.\n") ; break;
	case DDERR_OUTOFCAPS: debug.Output ("DirectDraw: The hardware needed for the requested operation has already been allocated.\n") ; break;
	case DDERR_OUTOFMEMORY: debug.Output ("DirectDraw: DirectDraw does not have enough memory to perform the operation.\n") ; break;
	case DDERR_OUTOFVIDEOMEMORY: debug.Output ("DirectDraw: DirectDraw does not have enough memory to perform the operation.\n") ; break;
	case DDERR_OVERLAYCANTCLIP: debug.Output ("DirectDraw: The hardware does not support clipped overlays.\n") ; break;
	case DDERR_OVERLAYCOLORKEYONLYONEACTIVE: debug.Output ("DirectDraw: Can only have ony color key active at one time for overlays.\n") ; break;
	case DDERR_OVERLAYNOTVISIBLE: debug.Output ("DirectDraw: Returned when GetOverlayPosition is called on a hidden overlay.\n") ; break;
	case DDERR_PALETTEBUSY: debug.Output ("DirectDraw: Access to this palette is being refused because the palette is already locked by another thread.\n") ; break;
	case DDERR_PRIMARYSURFACEALREADYEXISTS: debug.Output ("DirectDraw: This process already has created a primary surface.\n") ; break;
	case DDERR_REGIONTOOSMALL: debug.Output ("DirectDraw: Region passed to Clipper::GetClipList is too small.\n") ; break;
	case DDERR_SURFACEALREADYATTACHED: debug.Output ("DirectDraw: This surface is already attached to the surface it is being attached to.\n") ; break;
	case DDERR_SURFACEALREADYDEPENDENT: debug.Output ("DirectDraw: This surface is already a dependency of the surface it is being made a dependency of.\n") ; break;
	case DDERR_SURFACEBUSY: debug.Output ("DirectDraw: Access to this surface is being refused because the surface is locked.\nIf the problem is with the back buffer, try calling GameX.UnlockBackBuffer() first.\n") ; break;
	case DDERR_SURFACEISOBSCURED: debug.Output ("DirectDraw: Access to surface refused because the surface is obscured.\n") ; break;
	case DDERR_SURFACELOST: debug.Output ("DirectDraw: Access to this surface is being refused because the surface memory is gone. The DirectDrawSurface object representing this surface should have Restore called on it.\n") ; break;
	case DDERR_SURFACENOTATTACHED: debug.Output ("DirectDraw: The requested surface is not attached.\n") ; break;
	case DDERR_TOOBIGHEIGHT: debug.Output ("DirectDraw: Height requested by DirectDraw is too large.\n") ; break;
	case DDERR_TOOBIGSIZE: debug.Output ("DirectDraw: Size requested by DirectDraw is too large --	the individual height and width are OK.\n") ; break;
	case DDERR_TOOBIGWIDTH: debug.Output ("DirectDraw: Width requested by DirectDraw is too large.\n") ; break;
	case DDERR_UNSUPPORTED: debug.Output ("DirectDraw: Action not supported.\n") ; break;
	case DDERR_UNSUPPORTEDFORMAT: debug.Output ("DirectDraw: FOURCC format requested is unsupported by DirectDraw.\n") ; break;
	case DDERR_UNSUPPORTEDMASK: debug.Output ("DirectDraw: Bitmask in the pixel format requested is unsupported by DirectDraw.\n") ; break;
	case DDERR_VERTICALBLANKINPROGRESS: debug.Output ("DirectDraw: Vertical blank is in progress.\n") ; break;
	case DDERR_WASSTILLDRAWING: debug.Output ("DirectDraw: Informs DirectDraw that the previous Blt which is transfering information to or from this Surface is incomplete.\n") ; break;
	case DDERR_WRONGMODE: debug.Output ("DirectDraw: This surface can not be restored because it was created in a different mode.\n") ; break;
	default: debug.Output ("DirectDraw: Unknown Error.\n") ; dd_last_error = DD_OK; break;
	}
}

void WindowsDX::Direct3DError (HRESULT status)
{
	if(SUCCEEDED(status) || status == dd_last_error) return;
	dd_last_error = status;
	switch (status) {
	case D3DERR_WRONGTEXTUREFORMAT: debug.Output ("Direct3D: D3DERR_WRONGTEXTUREFORMAT.\n") ; break;
	case D3DERR_UNSUPPORTEDCOLOROPERATION: debug.Output ("Direct3D: D3DERR_UNSUPPORTEDCOLOROPERATION.\n") ; break;
	case D3DERR_UNSUPPORTEDCOLORARG: debug.Output ("Direct3D: D3DERR_UNSUPPORTEDCOLORARG.\n") ; break;
	case D3DERR_UNSUPPORTEDALPHAOPERATION: debug.Output ("Direct3D: D3DERR_UNSUPPORTEDALPHAOPERATION.\n") ; break;
	case D3DERR_UNSUPPORTEDALPHAARG: debug.Output ("Direct3D: D3DERR_UNSUPPORTEDALPHAARG.\n") ; break;
	case D3DERR_TOOMANYOPERATIONS: debug.Output ("Direct3D: D3DERR_TOOMANYOPERATIONS.\n") ; break;
	case D3DERR_CONFLICTINGTEXTUREFILTER: debug.Output ("Direct3D: D3DERR_CONFLICTINGTEXTUREFILTER.\n") ; break;
	case D3DERR_CONFLICTINGRENDERSTATE: debug.Output ("Direct3D: D3DERR_CONFLICTINGRENDERSTATE.\n") ; break;
	case D3DERR_UNSUPPORTEDTEXTUREFILTER: debug.Output ("Direct3D: D3DERR_UNSUPPORTEDTEXTUREFILTER.\n") ; break;
	case D3DERR_CONFLICTINGTEXTUREPALETTE: debug.Output ("Direct3D: D3DERR_CONFLICTINGTEXTUREPALETTE.\n") ; break;
	case E_NOINTERFACE: debug.Output ("Direct3D: E_NOINTERFACE.\n") ; break;
	default: debug.Output ("Direct3D: Unknown Error.\n") ; dd_last_error = DD_OK; break;
	}
}



void WindowsDX::ReportProblem (GameXStatus problem)
{
	win_status = problem;
	ReportProblem() ;
}

void WindowsDX::ReportProblem (void)
{	
	if(win_status_problem_ignore[win_status])
		win_status = GAMEX_READY;
	if(win_status == GAMEX_READY)
		return; // if there's no problem or it's already been reported and ignored

	win_status_problem_ignore[win_status] = true;

	char error[128] ;
	char recommend[640] ;	
	char disp[1280] ;
	
	switch (win_status) {
	case GAMEX_NO_INIT:
		sprintf (error, "Error: NO_INIT (User did not initialize GameX)") ;
		sprintf (recommend, "You must start GameX by calling:\n  GameX.Initialize (name, flags, xres, yres, speed)\n\nfrom the GameInit() function.\nFor example:\n  GameX.Initialize(\"My Game\",VIDEO_16BIT, 640, 480, 60);\n\nIf you are sure that you are making this call before anything else in GameInit(), make sure that you don't have any\nstatically constructed objects whose constructors use GameX in some way, such as by loading an image.\n\nSee documentation for details on starting GameX.") ;
		break;
	case GAMEX_INITCLASS_FAILED:
		sprintf (error, "Error: INITCLASS_FAILED (Cannot Initialize Window Class)") ;
		sprintf (recommend, "Are you using Windows 98/2000/NT/XP? GameX requires one of these. Still having trouble? Please report error and system configuration.") ;
		break;
	case GAMEX_INITWINTEMP_FAILED:
		sprintf (error, "Error: INITWINTEMP_FAILED (Cannot Initialize Temp Window)") ;
		sprintf (recommend, "Are you using Windows 98/2000/NT/XP? GameX requires one of these. Still having trouble? Please report error and system configuration.") ;
		break;
	case GAMEX_INITWINREAL_FAILED:
		sprintf (error, "Error: INITWINREAL_FAILED (Cannot Initialize Main Window)") ;
		sprintf (recommend, "Are you using Windows 98/2000/NT/XP? GameX requires one of these. Still having trouble? Please report error and system configuration.") ;
		break;
	case GAMEX_SHOWWIN_FAILED:
		sprintf (error, "Error: SHOWWIN_FAILED (Unable to show Main Window)") ;
		sprintf (recommend, "Are you using Windows 98/2000/NT/XP? GameX requires one of these. Still having trouble? Please report error and system configuration.") ;
		break;
	case GAMEX_DDSTART_FAILED:
		sprintf (error, "Error: DDSTART_FAILED (Unable to start DirectDraw or DirectX)") ;
		sprintf (recommend, "Have you installed DirectX 7.0 or later? GameX requires it. Still having trouble? Please report error and system configuration.") ;
		break;
	case GAMEX_D3DSTART_FAILED:
		sprintf (error, "Error: D3DSTART_FAILED (Unable to start Direct3D)") ;
		sprintf (recommend, "Have you installed DirectX 7.0 or later? GameX requires it. Still having trouble? Please report error and system configuration.") ;
		break;
	case GAMEX_DDSETLEVEL_FAILED:
		sprintf (error, "Error: DDSETLEVEL_FAILED (Unable to set DirectX Cooperative Level)") ;
		sprintf (recommend, "Have you installed DirectX 7.0 or later? GameX requires it. Still having trouble? Please report error and system configuration.") ;
		break;
	case GAMEX_SETMODE_FAILED:
		sprintf (error, "Error: SETMODE_FAILED (Unable to set screen mode)") ;
		sprintf (recommend, "Have you installed DirectX 7.0 or later? GameX requires it. Still having trouble? Please report error and system configuration.") ;
		break;
	case GAMEX_FRONTSURFACE_FAILED:
		sprintf (error, "Error: FRONTSURFACE_FAILED (Unable to create primary surface)") ;
		sprintf (recommend, "Have you installed DirectX 7.0 or later? GameX requires it. Still having trouble? Please report error and system configuration.") ;	
		break;
	case GAMEX_BACKSURFACE_FAILED:
		sprintf (error, "Error: BACKSURFACE_FAILED (Unable to create secondary surface)") ;
		sprintf (recommend, "Have you installed DirectX 7.0 or later? GameX requires it. Still having trouble? Please report error and system configuration.") ;
		break;
	case GAMEX_REQSIZE_FAILED:
		sprintf (error, "Error: REQSIZE_FAILED (Requested size invalid)") ;
		sprintf (recommend, "The requested game size was invalid.\nThe absolute maximum size allowed is 2048x2047, and the absolute minimum allowed is 64x64.\nTry to request a reasonable size though, such as something between 320x240 and 1600x1200.\nIf you are unsure of screen size, try initializing GameX as follows:\nGameX.Initialize (\"My Game\",VIDEO_16BIT, 800, 600);") ;
		break;
	case GAMEX_SOUNDINIT_FAILED:
		sprintf (error, "Error: SOUNDINIT_FAILED (Unable to initialize Direct Sound)") ;
		sprintf (recommend, "GameX was unable to initialize sound. Is sound card installed and working?") ;
		break;
	case GAMEX_SOUNDCOOP_FAILED:
		sprintf (error, "Error: SOUNDCOOP_FAILED (Unable to set Direct Sound Cooperative Level)") ;
		sprintf (recommend, "GameX was unable to set Direct Sound cooperative level. Is sound card installed and working?") ;
		break;	
	case GAMEX_SOUNDCREATE_FAILED:
		sprintf (error, "Error: SOUNDCREATE_FAILED (Unable to create Direct Sound buffer)") ;
		sprintf (recommend, "GameX was unable to play a sound, probably because it did not load correctly. Make sure to save your sounds in a standard format\n (8- or 16-bits, 11025- or 22050- or 44100- samples per second, and in PCM format).") ;
		break;
	case GAMEX_SOUNDMATCH_FAILED:
		sprintf (error, "Error: SOUNDMATCH_FAILED (Loaded sounds do not match)") ;
		sprintf (recommend, "You are attempting to load several sounds with different quality, resolution, and/or channel info. All sounds to be played simultaneously must have the same resolution (bits per sample), same quality (samples per second), and same number of channels (mono vs. stereo). Edit WAV files accordinging.") ;
		break;
	case GAMEX_SOUNDCAPS_FAILED:
		sprintf (error, "Error: SOUNDCAPS_FAILED (Unable to get Sound Buffer Capabilities)") ;
		sprintf (recommend, "GameX was unable to retrieve sound buffer capabilities. You may be running out of memory. Close any open applications. Make sure you have DirectX 7.0 or later.") ;
		break;
	case GAMEX_SOUNDBUFLEN_FAILED:
		sprintf (error, "Error: SOUNDBUFLEN_FAILED (Direct Sound buffer and SoundX buffers do not match)") ;
		sprintf (recommend, "Direct Sound and SoundX buffer lengths do not match. Internal error. Please report this problem and the code that generated it.") ;	
		break;
	case GAMEX_SOUNDLOST_FAIL:
		sprintf (error, "Error: SOUNDLOST_FAIL (Direct Sound buffer was lost)") ;
		sprintf (recommend, "Direct Sound buffer was lost. Make sure you are using DirectX 7.0 or later. Close any open applications. Try restarting system. If you are using a large or high quality sound file, try reducing it.") ;
		break;
	case GAMEX_SOUNDINVALID_FAIL:
		sprintf (error, "Error: SOUNDINVALID_FAIL (Direct Sound Invalid Call)") ;
		sprintf (recommend, "DirectX, via GameX, made an invalid sound call. Make sure you are using DirectX 7.0 or later. Close any open applications. Try restarting system. If you are using a large or high quality sound file, try reducing it.") ;
		break;
	case GAMEX_SOUNDPARAM_FAIL:
		sprintf (error, "Error: SOUNDPARAM_FAIL (Direct Sound Invalid Parameter)") ;
		sprintf (recommend, "DirectX, via GameX, used an invalid parameter. Make sure you are using DirectX 7.0 or later. Close any open applications. Try restarting system. If you are using a large or high quality sound file, try reducing it.") ;
		break;
	case GAMEX_SOUNDPRIOLEVEL_FAIL:
		sprintf (error, "Error: SOUNDPARAM_FAIL (Direct Sound needs Priority Level)") ;
		sprintf (recommend, "DirectX, via GameX, is using an invalid priority level. Make sure you are using DirectX 7.0 or later. Close any open applications. Try restarting system. If you are using a large or high quality sound file, try reducing it.") ;		
		break;	
	case GAMEX_UNINIT_IMG_DRAW_FAIL:
		sprintf (error, "Error: UNINIT_IMG_DRAW_FAIL (Attempted to draw an ImageX that was not ready to be drawn)") ;
		sprintf (recommend,"You must Load() or Create() each ImageX before it can be drawn.\nAlso, make sure you do not delete, Destroy(), or\nallow an ImageX to go out of scope while it is in use.\n(When an image is drawn, it can remain in use until GameDraw() returns.)");
		break;
	case GAMEX_INVALID_DRAW_TIME_FAIL:
		sprintf (error, "Error: INVALID_DRAW_TIME_FAIL (Attempted to draw an ImageX or other graphic to the screen from GameRun())") ;
		sprintf (recommend,"You cannot use the drawing functions to draw to the screen in GameRun() or in ANY functions that GameRun() calls.\n(Or in anything they in turn call, etc.)\nThese drawing functions should only be called via GameDraw(), not GameRun().\nKeep in mind that the purpose of GameRun() is to update the game state without drawing it,\nand the purpose of GameDraw() is to draw the game state without updating it.");
		break;
	case GAMEX_PIXEL_ACCESS_FAILURE:
		sprintf (error, "Error: PIXEL_ACCESS_FAILURE (Attempted to call DrawPixel or ReadPixel without first calling AccessPixels())") ;
 		sprintf (recommend,"You must call GameX.AccessPixels(&image) immediately before directly drawing to or reading from an image's pixels,\nand you must call GameX.EndPixelAccess() when done (and before calling any other drawing functions).\n\nIf you want to draw pixels directly to the screen, the access call is GameX.AccessPixels(VIEWPORT).\n");
		break;
	case GAMEX_CLIPPER_FAILED:
		sprintf (error, "Error: CLIPPER_FAILED (Failed to create or set a clipper for the back buffer)") ;
		sprintf (recommend, "Are you using DirectX 7.0 or later? GameX requires one of these. Still having trouble? Please report error and system configuration.") ;
		break;
	default:
		sprintf (error, "Error: UNKNOWN") ;
		sprintf (recommend, "Unregistered error with GameX. Please report. (Unless you caused this error by editing GameX)") ;
		break;
	}
	sprintf (disp, "GameX %s\n", error) ;
	debug.Output (disp) ; 
	int choice;

	if(win_status == GAMEX_DDSTART_FAILED) {
		sprintf (disp, "GameX has encountered a fatal error.\n\n%s\n\n%s\n\nIf restarting does not fix this problem, please (re)install DirectX and try again.", error, recommend) ;
		MessageBox (win_hwnd, disp, "GameX Error", MB_ICONSTOP) ;
		choice = IDNO;
	} else {
		sprintf (disp, "GameX has encountered an error and may not be able to continue.\n\n%s\n\n%s\n\nAttempt to continue execution anyway?", error, recommend) ;
		if(dd_main != NULL)
			win_status = GAMEX_READY;
		choice = MessageBox (win_hwnd, disp, "GameX Error", MB_YESNO|MB_DEFBUTTON2|MB_ICONSTOP) ;
	}
	if(choice == IDNO)
		Quit(true);
	else if (win_status == GAMEX_NO_INIT)
		Initialize("Improperly Initialized Game", VIDEO_16BIT|VIDEO_WINDOWED,640,480);

} // insert breakpoint here :-)
